package ru.dbotthepony.kstarbound.world.api

import ru.dbotthepony.kstarbound.RegistryObject
import ru.dbotthepony.kstarbound.defs.tile.LiquidDefinition
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import java.io.DataInputStream

interface IChunkCell {
	val foreground: ITileState
	val background: ITileState
	val liquid: ILiquidState

	var dungeonId: Int
	var biome: Int
	var envBiome: Int
	var isIndestructible: Boolean

	val isEmpty: Boolean

	fun read(
		materialAccess: (Int) -> RegistryObject<TileDefinition>?,
		modifierAccess: (Int) -> RegistryObject<MaterialModifier>?,
		liquidAccess: (Int) -> RegistryObject<LiquidDefinition>?,
		stream: DataInputStream
	) {
		foreground.read(materialAccess, modifierAccess, stream)
		background.read(materialAccess, modifierAccess, stream)
		liquid.read(liquidAccess, stream)

		stream.skipBytes(1) // collisionMap

		dungeonId = stream.readUnsignedShort()
		biome = stream.readUnsignedByte()
		envBiome = stream.readUnsignedByte()
		isIndestructible = stream.readBoolean()

		stream.skipBytes(1) // unknown
	}

	companion object : IChunkCell {
		override val foreground: ITileState
			get() = ITileState
		override val background: ITileState
			get() = ITileState
		override val liquid: ILiquidState
			get() = ILiquidState
		override var dungeonId: Int
			get() = 0
			set(value) { throw UnsupportedOperationException("Empty chunk cell") }
		override var biome: Int
			get() = 0
			set(value) { throw UnsupportedOperationException("Empty chunk cell") }
		override var envBiome: Int
			get() = 0
			set(value) { throw UnsupportedOperationException("Empty chunk cell") }
		override var isIndestructible: Boolean
			get() = false
			set(value) { throw UnsupportedOperationException("Empty chunk cell") }
		override val isEmpty: Boolean
			get() = true
	}
}